import unittest
from textwrap import dedent

from rope.base import builtins, libutils, pyobjects
from rope.base.builtins import Dict
from ropetest import testutils


class BuiltinTypesTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_case(self):
        self.mod.write("l = []\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue("append" in pymod["l"].get_object())

    def test_holding_type_information(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            a_var = l.pop()
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_get_items(self):
        self.mod.write(dedent("""\
            class C(object):
                def __getitem__(self, i):
                    return C()
            c = C()
            a_var = c[0]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_get_items_for_lists(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            a_var = l[0]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_get_items_from_slices(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            a_var = l[:].pop()
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_simple_for_loops(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            for c in l:
                a_var = c
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_definition_location_for_loop_variables(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            for c in l:
                pass
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_var = pymod["c"]
        self.assertEqual((pymod, 4), c_var.get_definition_location())

    def test_simple_case_for_dicts(self):
        self.mod.write("d = {}\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue("get" in pymod["d"].get_object())

    def test_get_item_for_dicts(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            d = {1: C()}
            a_var = d[1]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_dict_function_parent(self):
        self.mod.write(dedent("""\
            d = {1: 2}
            a_var = d.keys()
        """))
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod["d"].get_object()["keys"].get_object()
        self.assertEqual(type(a_var.parent), Dict)

    def test_popping_dicts(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            d = {1: C()}
            a_var = d.pop(1)
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_getting_keys_from_dicts(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {C1(): C2()}
            for c in d.keys():
                a_var = c
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C1"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_getting_values_from_dicts(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {C1(): C2()}
            for c in d.values():
                a_var = c
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C2"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_getting_iterkeys_from_dicts(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {C1(): C2()}
            for c in d.keys():
                a_var = c
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C1"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_getting_itervalues_from_dicts(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {C1(): C2()}
            for c in d.values():
                a_var = c
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C2"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_using_copy_for_dicts(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {C1(): C2()}
            for c in d.copy():
                a_var = c
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C1"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_tuple_assignments_for_items(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {C1(): C2()}
            key, value = d.items()[0]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        key = pymod["key"].get_object()
        value = pymod["value"].get_object()
        self.assertEqual(c1_class, key.get_type())
        self.assertEqual(c2_class, value.get_type())

    def test_tuple_assignment_for_lists(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C(), C()]
            a, b = l
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c_class, a_var.get_type())
        self.assertEqual(c_class, b_var.get_type())

    def test_tuple_assignments_for_iteritems_in_fors(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {C1(): C2()}
            for x, y in d.items():
                a = x;
                b = y
        """))
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_simple_tuple_assignments(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            a, b = C1(), C2()
        """))
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_overriding_builtin_names(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            list = C
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        list_var = pymod["list"].get_object()
        self.assertEqual(c_class, list_var)

    def test_simple_builtin_scope_test(self):
        self.mod.write("l = list()\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue("append" in pymod["l"].get_object())

    def test_simple_sets(self):
        self.mod.write("s = set()\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue("add" in pymod["s"].get_object())

    def test_making_lists_using_the_passed_argument_to_init(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l1 = [C()]
            l2 = list(l1)
            a_var = l2.pop()
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_making_tuples_using_the_passed_argument_to_init(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l1 = [C()]
            l2 = tuple(l1)
            a_var = l2[0]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_making_sets_using_the_passed_argument_to_init(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l1 = [C()]
            l2 = set(l1)
            a_var = l2.pop()
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_making_dicts_using_the_passed_argument_to_init(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            l1 = [(C1(), C2())]
            l2 = dict(l1)
            a, b = l2.items()[0]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_range_builtin_function(self):
        self.mod.write("l = range(1)\n")
        pymod = self.project.get_pymodule(self.mod)
        l = pymod["l"].get_object()
        self.assertTrue("append" in l)

    def test_reversed_builtin_function(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            for x in reversed(l):
                a_var = x
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_sorted_builtin_function(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            a_var = sorted(l).pop()
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_super_builtin_function(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            class A(object):
                def a_f(self):
                    return C()
            class B(A):
                def b_f(self):
                    return super(B, self).a_f()
            a_var = B.b_f()
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_file_builtin_type(self):
        self.mod.write(dedent("""\
            for line in open("file.txt"):
                a_var = line
        """))
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod["a_var"].get_object()
        self.assertTrue(isinstance(a_var.get_type(), builtins.Str))

    def test_property_builtin_type(self):
        self.mod.write("p = property()\n")
        pymod = self.project.get_pymodule(self.mod)
        p_var = pymod["p"].get_object()
        self.assertTrue("fget" in p_var)

    def test_lambda_functions(self):
        self.mod.write("l = lambda: 1\n")
        pymod = self.project.get_pymodule(self.mod)
        l_var = pymod["l"].get_object()
        self.assertEqual(pyobjects.get_base_type("Function"), l_var.get_type())

    def test_lambda_function_definition(self):
        self.mod.write("l = lambda x, y = 2, *a, **b: x + y\n")
        pymod = self.project.get_pymodule(self.mod)
        l_var = pymod["l"].get_object()
        self.assertTrue(l_var.get_name() is not None)
        self.assertEqual(len(l_var.get_param_names()), 4)
        self.assertEqual((pymod, 1), pymod["l"].get_definition_location())

    def test_lambdas_that_return_unknown(self):
        self.mod.write("a_var = (lambda: None)()\n")
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod["a_var"].get_object()
        self.assertTrue(a_var is not None)

    def test_builtin_zip_function(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            c1_list = [C1()]
            c2_list = [C2()]
            a, b = zip(c1_list, c2_list)[0]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_builtin_zip_function_with_more_than_two_args(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            c1_list = [C1()]
            c2_list = [C2()]
            a, b, c = zip(c1_list, c2_list, c1_list)[0]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        c_var = pymod["c"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())
        self.assertEqual(c1_class, c_var.get_type())

    def test_wrong_arguments_to_zip_function(self):
        self.mod.write(dedent("""\
            class C1(object):
                pass
            c1_list = [C1()]
            a, b = zip(c1_list, 1)[0]
        """))
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()  # noqa
        self.assertEqual(c1_class, a_var.get_type())

    def test_enumerate_builtin_function(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            for i, x in enumerate(l):
                a_var = x
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_builtin_class_get_name(self):
        self.assertEqual("object", builtins.builtins["object"].get_object().get_name())
        self.assertEqual(
            "property", builtins.builtins["property"].get_object().get_name()
        )

    def test_star_args_and_double_star_args(self):
        self.mod.write(dedent("""\
            def func(p, *args, **kwds):
                pass
        """))
        pymod = self.project.get_pymodule(self.mod)
        func_scope = pymod["func"].get_object().get_scope()
        args = func_scope["args"].get_object()
        kwds = func_scope["kwds"].get_object()
        self.assertTrue(isinstance(args.get_type(), builtins.List))
        self.assertTrue(isinstance(kwds.get_type(), builtins.Dict))

    def test_simple_list_comprehension_test(self):
        self.mod.write("a_var = [i for i in range(10)]\n")
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod["a_var"].get_object()
        self.assertTrue(isinstance(a_var.get_type(), builtins.List))

    def test_simple_list_generator_expression(self):
        self.mod.write("a_var = (i for i in range(10))\n")
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod["a_var"].get_object()
        self.assertTrue(isinstance(a_var.get_type(), builtins.Iterator))

    def test_iter_builtin_function(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = [C()]
            for c in iter(l):
                a_var = c
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_simple_int_type(self):
        self.mod.write("l = 1\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            builtins.builtins["int"].get_object(), pymod["l"].get_object().get_type()
        )

    def test_simple_float_type(self):
        self.mod.write("l = 1.0\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            builtins.builtins["float"].get_object(), pymod["l"].get_object().get_type()
        )

    def test_simple_float_type2(self):
        self.mod.write("l = 1e1\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            builtins.builtins["float"].get_object(), pymod["l"].get_object().get_type()
        )

    def test_simple_complex_type(self):
        self.mod.write("l = 1.0j\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            builtins.builtins["complex"].get_object(),
            pymod["l"].get_object().get_type(),
        )

    def test_handling_unaryop_on_ints(self):
        self.mod.write("l = -(1)\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            builtins.builtins["int"].get_object(), pymod["l"].get_object().get_type()
        )

    def test_handling_binop_on_ints(self):
        self.mod.write("l = 1 + 1\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            builtins.builtins["int"].get_object(), pymod["l"].get_object().get_type()
        )

    def test_handling_compares(self):
        self.mod.write("l = 1 == 1\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            builtins.builtins["bool"].get_object(), pymod["l"].get_object().get_type()
        )

    def test_handling_boolops(self):
        self.mod.write("l = 1 and 2\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            builtins.builtins["int"].get_object(), pymod["l"].get_object().get_type()
        )

    def test_binary_or_left_value_unknown(self):
        code = "var = (asdsd or 3)\n"
        pymod = libutils.get_string_module(self.project, code)
        self.assertEqual(
            builtins.builtins["int"].get_object(), pymod["var"].get_object().get_type()
        )

    def test_unknown_return_object(self):
        src = dedent("""\
            import sys
            def foo():
              res = set(sys.builtin_module_names)
              if foo: res.add(bar)
        """)
        self.project.prefs["import_dynload_stdmods"] = True
        self.mod.write(src)
        self.project.pycore.analyze_module(self.mod)

    def test_abstractmethods_attribute(self):
        # see http://bugs.python.org/issue10006 for details
        src = "class SubType(type): pass\nsubtype = SubType()\n"
        self.mod.write(src)
        self.project.pycore.analyze_module(self.mod)


class BuiltinModulesTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project(
            extension_modules=["time", "invalid", "invalid.sub"]
        )
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_case(self):
        self.mod.write("import time")
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue("time" in pymod["time"].get_object())

    def test_ignored_extensions(self):
        self.mod.write("import os")
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue("rename" not in pymod["os"].get_object())

    def test_ignored_extensions_2(self):
        self.mod.write("import os")
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue("rename" not in pymod["os"].get_object())

    def test_nonexistent_modules(self):
        self.mod.write("import invalid")
        pymod = self.project.get_pymodule(self.mod)
        pymod["invalid"].get_object()

    def test_nonexistent_modules_2(self):
        self.mod.write(dedent("""\
            import invalid
            import invalid.sub
        """))
        pymod = self.project.get_pymodule(self.mod)
        invalid = pymod["invalid"].get_object()
        self.assertTrue("sub" in invalid)

    def test_time_in_std_mods(self):
        import rope.base.stdmods

        self.assertTrue("time" in rope.base.stdmods.standard_modules())

    def test_timemodule_normalizes_to_time(self):
        import rope.base.stdmods

        self.assertEqual(rope.base.stdmods.normalize_so_name("timemodule.so"), "time")



import re
from contextlib import contextmanager
from itertools import chain
from typing import Dict

from rope.base import ast, codeanalyze
from rope.base.change import ChangeContents, ChangeSet
from rope.base.exceptions import RefactoringError
from rope.base.utils.datastructures import OrderedSet
from rope.refactor import patchedast, similarfinder, sourceutils, suites, usefunction


# Extract refactoring has lots of special cases.  I tried to split it
# to smaller parts to make it more manageable:
#
# _ExtractInfo: holds information about the refactoring; it is passed
# to the parts that need to have information about the refactoring
#
# _ExtractCollector: merely saves all of the information necessary for
# performing the refactoring.
#
# _DefinitionLocationFinder: finds where to insert the definition.
#
# _ExceptionalConditionChecker: checks for exceptional conditions in
# which the refactoring cannot be applied.
#
# _ExtractMethodParts: generates the pieces of code (like definition)
# needed for performing extract method.
#
# _ExtractVariableParts: like _ExtractMethodParts for variables.
#
# _ExtractPerformer: Uses above classes to collect refactoring
# changes.
#
# There are a few more helper functions and classes used by above
# classes.
class _ExtractRefactoring:
    kind_prefixes: Dict[str, str] = {}

    def __init__(self, project, resource, start_offset, end_offset, variable=False):
        self.project = project
        self.resource = resource
        self.start_offset = self._fix_start(resource.read(), start_offset)
        self.end_offset = self._fix_end(resource.read(), end_offset)

    def _fix_start(self, source, offset):
        while offset < len(source) and source[offset].isspace():
            offset += 1
        return offset

    def _fix_end(self, source, offset):
        while offset > 0 and source[offset - 1].isspace():
            offset -= 1
        return offset

    def get_changes(self, extracted_name, similar=False, global_=False, kind=None):
        """Get the changes this refactoring makes

        :parameters:
            - `extracted_name`: target name, when starts with @ - set kind to
            classmethod, $ - staticmethod
            - `similar`: if `True`, similar expressions/statements are also
              replaced.
            - `global_`: if `True`, the extracted method/variable will
              be global.
            - `kind`: kind of target refactoring to (staticmethod, classmethod)

        """
        extracted_name, kind = self._get_kind_from_name(extracted_name, kind)

        info = _ExtractInfo(
            self.project,
            self.resource,
            self.start_offset,
            self.end_offset,
            extracted_name,
            variable=self._get_kind(kind) == "variable",
            similar=similar,
            make_global=global_,
        )
        info.kind = self._get_kind(kind)
        new_contents = _ExtractPerformer(info).extract()
        changes = ChangeSet(f"Extract {info.kind} <{extracted_name}>")
        changes.add_change(ChangeContents(self.resource, new_contents))
        return changes

    def _get_kind_from_name(self, extracted_name, kind):
        for sign, selected_kind in self.kind_prefixes.items():
            if extracted_name.startswith(sign):
                self._validate_kind_prefix(kind, selected_kind)
                return extracted_name[1:], selected_kind
        return extracted_name, kind

    @staticmethod
    def _validate_kind_prefix(kind, selected_kind):
        if kind and kind != selected_kind:
            raise RefactoringError("Kind and shortcut in name mismatch")

    @classmethod
    def _get_kind(cls, kind):
        raise NotImplementedError(f"You have to subclass {cls}")


class ExtractMethod(_ExtractRefactoring):
    kind = "method"
    allowed_kinds = ("function", "method", "staticmethod", "classmethod")
    kind_prefixes = {"@": "classmethod", "$": "staticmethod"}

    @classmethod
    def _get_kind(cls, kind):
        return kind if kind in cls.allowed_kinds else cls.kind


class ExtractVariable(_ExtractRefactoring):
    def __init__(self, *args, **kwds):
        kwds = dict(kwds)
        kwds["variable"] = True
        super().__init__(*args, **kwds)

    kind = "variable"

    def _get_kind(cls, kind):
        return cls.kind


class _ExtractInfo:
    """Holds information about the extract to be performed"""

    def __init__(
        self, project, resource, start, end, new_name, variable, similar, make_global
    ):
        self.project = project
        self.resource = resource
        self.pymodule = project.get_pymodule(resource)
        self.global_scope = self.pymodule.get_scope()
        self.source = self.pymodule.source_code
        self.lines = self.pymodule.lines
        self.new_name = new_name
        self.variable = variable
        self.similar = similar
        self._init_parts(start, end)
        self.kind = None
        self._init_scope()
        self.make_global = make_global

    def _init_parts(self, start, end):
        self.region = (
            self._choose_closest_line_end(start),
            self._choose_closest_line_end(end, end=True),
        )

        start = self.logical_lines.logical_line_in(
            self.lines.get_line_number(self.region[0])
        )[0]
        end = self.logical_lines.logical_line_in(
            self.lines.get_line_number(self.region[1])
        )[1]
        self.region_lines = (start, end)

        self.lines_region = (
            self.lines.get_line_start(self.region_lines[0]),
            self.lines.get_line_end(self.region_lines[1]),
        )

    @property
    def logical_lines(self):
        return self.pymodule.logical_lines

    def _init_scope(self):
        start_line = self.region_lines[0]
        scope = self.global_scope.get_inner_scope_for_line(start_line)
        if scope.get_kind() != "Module" and scope.get_start() == start_line:
            scope = scope.parent
        self.scope = scope
        self.scope_region = self._get_scope_region(self.scope)

    def _get_scope_region(self, scope):
        return (
            self.lines.get_line_start(scope.get_start()),
            self.lines.get_line_end(scope.get_end()) + 1,
        )

    def _choose_closest_line_end(self, offset, end=False):
        lineno = self.lines.get_line_number(offset)
        line_start = self.lines.get_line_start(lineno)
        line_end = self.lines.get_line_end(lineno)
        if self.source[line_start:offset].strip() == "":
            if end:
                return line_start - 1
            else:
                return line_start
        elif self.source[offset:line_end].strip() == "":
            return min(line_end, len(self.source))
        return offset

    @property
    def one_line(self):
        return self.region != self.lines_region and (
            self.logical_lines.logical_line_in(self.region_lines[0])
            == self.logical_lines.logical_line_in(self.region_lines[1])
        )

    @property
    def global_(self):
        return self.scope.parent is None

    @property
    def method(self):
        return self.scope.parent is not None and self.scope.parent.get_kind() == "Class"

    @property
    def indents(self):
        return sourceutils.get_indents(self.pymodule.lines, self.region_lines[0])

    @property
    def scope_indents(self):
        if self.global_:
            return 0
        return sourceutils.get_indents(self.pymodule.lines, self.scope.get_start())

    @property
    def extracted(self):
        return self.source[self.region[0] : self.region[1]]

    _cached_parsed_extracted = None

    @property
    def _parsed_extracted(self):
        if self._cached_parsed_extracted is None:
            self._cached_parsed_extracted = _parse_text(self.extracted)
        return self._cached_parsed_extracted

    _returned = None

    @property
    def returned(self):
        """Does the extracted piece contain return statement"""
        if self._returned is None:
            self._returned = usefunction._returns_last(self._parsed_extracted)
        return self._returned

    _returning_named_expr = None

    @property
    def returning_named_expr(self):
        """Does the extracted piece contains named expression/:= operator)"""
        if self._returning_named_expr is None:
            self._returning_named_expr = usefunction._namedexpr_last(
                self._parsed_extracted
            )
        return self._returning_named_expr

    _returning_generator = None

    @property
    def returning_generator_exp(self):
        """Does the extracted piece contains a generator expression"""
        if self._returning_generator is None:
            self._returning_generator = (
                isinstance(self._parsed_extracted, ast.Module)
                and isinstance(self._parsed_extracted.body[0], ast.Expr)
                and isinstance(self._parsed_extracted.body[0].value, ast.GeneratorExp)
            )

        return self._returning_generator


class _ExtractCollector:
    """Collects information needed for performing the extract"""

    def __init__(self, info):
        self.definition = None
        self.body_pattern = None
        self.checks = {}
        self.replacement_pattern = None
        self.matches = None
        self.replacements = None
        self.definition_location = None


class _ExtractPerformer:
    def __init__(self, info):
        self.info = info
        _ExceptionalConditionChecker()(self.info)

    def extract(self):
        extract_info = self._collect_info()
        content = codeanalyze.ChangeCollector(self.info.source)
        definition = extract_info.definition
        lineno, indents = extract_info.definition_location
        offset = self.info.lines.get_line_start(lineno)
        indented = sourceutils.fix_indentation(definition, indents)
        content.add_change(offset, offset, indented)
        self._replace_occurrences(content, extract_info)
        return content.get_changed()

    def _replace_occurrences(self, content, extract_info):
        for match in extract_info.matches:
            replacement = similarfinder.CodeTemplate(extract_info.replacement_pattern)
            mapping = {}
            for name in replacement.get_names():
                node = match.get_ast(name)
                if node:
                    start, end = patchedast.node_region(match.get_ast(name))
                    mapping[name] = self.info.source[start:end]
                else:
                    mapping[name] = name
            region = match.get_region()
            content.add_change(region[0], region[1], replacement.substitute(mapping))

    def _collect_info(self):
        extract_collector = _ExtractCollector(self.info)
        self._find_definition(extract_collector)
        self._find_matches(extract_collector)
        self._find_definition_location(extract_collector)
        return extract_collector

    def _find_matches(self, collector):
        regions = self._where_to_search()
        finder = similarfinder.SimilarFinder(self.info.pymodule)
        matches = []
        for start, end in regions:
            region_matches = finder.get_matches(
                collector.body_pattern, collector.checks, start, end
            )
            # Don't extract overlapping regions
            last_match_end = -1
            for region_match in region_matches:
                if self.info.one_line and self._is_assignment(region_match):
                    continue
                start, end = region_match.get_region()
                if last_match_end < start:
                    matches.append(region_match)
                    last_match_end = end
        collector.matches = matches

    @staticmethod
    def _is_assignment(region_match):
        return isinstance(
            region_match.ast, (ast.Attribute, ast.Subscript)
        ) and isinstance(region_match.ast.ctx, ast.Store)

    def _where_to_search(self):
        if self.info.similar:
            if self.info.make_global or self.info.global_:
                return [(0, len(self.info.pymodule.source_code))]
            if self.info.method and not self.info.variable:
                class_scope = self.info.scope.parent
                regions = []
                method_kind = _get_function_kind(self.info.scope)
                for scope in class_scope.get_scopes():
                    if (
                        method_kind == "method"
                        and _get_function_kind(scope) != "method"
                    ):
                        continue
                    start = self.info.lines.get_line_start(scope.get_start())
                    end = self.info.lines.get_line_end(scope.get_end())
                    regions.append((start, end))
                return regions
            else:
                if self.info.variable:
                    return [self.info.scope_region]
                else:
                    return [self.info._get_scope_region(self.info.scope.parent)]
        else:
            return [self.info.region]

    def _find_definition_location(self, collector):
        matched_lines = []
        for match in collector.matches:
            start = self.info.lines.get_line_number(match.get_region()[0])
            start_line = self.info.logical_lines.logical_line_in(start)[0]
            matched_lines.append(start_line)
        location_finder = _DefinitionLocationFinder(self.info, matched_lines)
        collector.definition_location = (
            location_finder.find_lineno(),
            location_finder.find_indents(),
        )

    def _find_definition(self, collector):
        if self.info.variable:
            parts = _ExtractVariableParts(self.info)
        else:
            parts = _ExtractMethodParts(self.info)
        collector.definition = parts.get_definition()
        collector.body_pattern = parts.get_body_pattern()
        collector.replacement_pattern = parts.get_replacement_pattern()
        collector.checks = parts.get_checks()


class _DefinitionLocationFinder:
    def __init__(self, info, matched_lines):
        self.info = info
        self.matched_lines = matched_lines
        # This only happens when subexpressions cannot be matched
        if not matched_lines:
            self.matched_lines.append(self.info.region_lines[0])

    def find_lineno(self):
        if self.info.variable and not self.info.make_global:
            return self._get_before_line()
        if self.info.global_:
            toplevel = self._find_toplevel(self.info.scope)
            ast = self.info.pymodule.get_ast()
            newlines = sorted(self.matched_lines + [toplevel.get_end() + 1])
            return suites.find_visible(ast, newlines)
        if self.info.make_global:
            toplevel = self._find_toplevel(self.info.scope)
            return toplevel.get_end() + 1
        return self._get_after_scope()

    def _find_toplevel(self, scope):
        toplevel = scope
        if toplevel.parent is not None:
            while toplevel.parent.parent is not None:
                toplevel = toplevel.parent
        return toplevel

    def find_indents(self):
        if self.info.variable and not self.info.make_global:
            return sourceutils.get_indents(self.info.lines, self._get_before_line())
        else:
            if self.info.global_ or self.info.make_global:
                return 0
        return self.info.scope_indents

    def _get_before_line(self):
        ast = self.info.scope.pyobject.get_ast()
        return suites.find_visible(ast, self.matched_lines)

    def _get_after_scope(self):
        return self.info.scope.get_end() + 1


class _ExceptionalConditionChecker:
    def __call__(self, info):
        self.base_conditions(info)
        if info.one_line:
            self.one_line_conditions(info)
        else:
            self.multi_line_conditions(info)

    def base_conditions(self, info):
        if info.region[1] > info.scope_region[1]:
            raise RefactoringError("Bad region selected for extract method")

        end_line = info.region_lines[1]
        end_scope = info.global_scope.get_inner_scope_for_line(end_line)

        if end_scope != info.scope and end_scope.get_end() != end_line:
            raise RefactoringError("Bad region selected for extract method")
        try:
            extracted = info.extracted
            if info.one_line:
                extracted = "(%s)" % extracted
            if _UnmatchedBreakOrContinueFinder.has_errors(extracted):
                raise RefactoringError(
                    "A break/continue without having a matching for/while loop."
                )
        except SyntaxError:
            raise RefactoringError(
                "Extracted piece should contain complete statements."
            )

    def one_line_conditions(self, info):
        if self._is_region_on_a_word(info):
            raise RefactoringError("Should extract complete statements.")
        if info.variable and not info.one_line:
            raise RefactoringError("Extract variable should not span multiple lines.")
        if usefunction._named_expr_count(
            info._parsed_extracted
        ) - usefunction._namedexpr_last(info._parsed_extracted):
            raise RefactoringError(
                "Extracted piece cannot contain named expression (:= operator)."
            )

    def multi_line_conditions(self, info):
        node = _parse_text(info.source[info.region[0] : info.region[1]])
        count = usefunction._return_count(node)
        extracted = info.extracted
        if count > 1:
            raise RefactoringError(
                "Extracted piece can have only one return statement."
            )
        if usefunction._yield_count(node):
            raise RefactoringError("Extracted piece cannot have yield statements.")
        if not hasattr(
            ast, "PyCF_ALLOW_TOP_LEVEL_AWAIT"
        ) and _AsyncStatementFinder.has_errors(extracted):
            raise RefactoringError(
                "Extracted piece can only have async/await "
                "statements if Rope is running on Python "
                "3.8 or higher"
            )
        if count == 1 and not usefunction._returns_last(node):
            raise RefactoringError("Return should be the last statement.")
        if info.region != info.lines_region:
            raise RefactoringError(
                "Extracted piece should contain complete statements."
            )
        unbalanced_region_finder = _UnbalancedRegionFinder(
            info.region_lines[0], info.region_lines[1]
        )
        unbalanced_region_finder.visit(info.pymodule.ast_node)
        if unbalanced_region_finder.error:
            raise RefactoringError(
                "Extracted piece cannot contain the start of a block without the end."
            )

    def _is_region_on_a_word(self, info):
        if (
            info.region[0] > 0
            and self._is_on_a_word(info, info.region[0] - 1)
            or self._is_on_a_word(info, info.region[1] - 1)
        ):
            return True

    def _is_on_a_word(self, info, offset):
        prev = info.source[offset]
        if not (prev.isalnum() or prev == "_") or offset + 1 == len(info.source):
            return False
        next = info.source[offset + 1]
        return next.isalnum() or next == "_"


class _ExtractMethodParts(ast.RopeNodeVisitor):
    def __init__(self, info):
        self.info = info
        self.info_collector = self._create_info_collector()
        self.info.kind = self._get_kind_by_scope()
        self._check_constraints()

    def _get_kind_by_scope(self):
        if self._extacting_from_staticmethod():
            return "staticmethod"
        elif self._extracting_from_classmethod():
            return "classmethod"
        return self.info.kind

    def _check_constraints(self):
        if self._extracting_staticmethod() or self._extracting_classmethod():
            if not self.info.method:
                raise RefactoringError(
                    "Cannot extract to staticmethod/classmethod outside class"
                )

    def _extacting_from_staticmethod(self):
        return (
            self.info.method and _get_function_kind(self.info.scope) == "staticmethod"
        )

    def _extracting_from_classmethod(self):
        return self.info.method and _get_function_kind(self.info.scope) == "classmethod"

    def get_definition(self):
        if self.info.global_:
            return "\n%s\n" % self._get_function_definition()
        else:
            return "\n%s" % self._get_function_definition()

    def get_replacement_pattern(self):
        variables = []
        variables.extend(self._find_function_arguments())
        variables.extend(self._find_function_returns())
        return similarfinder.make_pattern(self._get_call(), variables)

    def get_body_pattern(self):
        variables = []
        variables.extend(self._find_function_arguments())
        variables.extend(self._find_function_returns())
        variables.extend(self._find_temps())
        return similarfinder.make_pattern(self._get_body(), variables)

    def _get_body(self):
        result = sourceutils.fix_indentation(self.info.extracted, 0)
        if self.info.one_line:
            result = "(%s)" % result
        return result

    def _find_temps(self):
        return usefunction.find_temps(self.info.project, self._get_body())

    def get_checks(self):
        if self.info.method and not self.info.make_global:
            if _get_function_kind(self.info.scope) == "method":
                class_name = similarfinder._pydefined_to_str(
                    self.info.scope.parent.pyobject
                )
                return {self._get_self_name(): "type=" + class_name}
        return {}

    def _create_info_collector(self):
        zero = self.info.scope.get_start() - 1
        start_line = self.info.region_lines[0] - zero
        end_line = self.info.region_lines[1] - zero
        info_collector = _FunctionInformationCollector(
            start_line, end_line, self.info.global_
        )
        body = self.info.source[self.info.scope_region[0] : self.info.scope_region[1]]
        node = _parse_text(body)
        info_collector.visit(node)
        return info_collector

    def _get_function_definition(self):
        args = self._find_function_arguments()
        returns = self._find_function_returns()

        result = []
        self._append_decorators(result)
        result.append("def %s:\n" % self._get_function_signature(args))
        unindented_body = self._get_unindented_function_body(returns)
        indents = sourceutils.get_indent(self.info.project)
        function_body = sourceutils.indent_lines(unindented_body, indents)
        result.append(function_body)
        definition = "".join(result)

        return definition + "\n"

    def _append_decorators(self, result):
        if self._extracting_staticmethod():
            result.append("@staticmethod\n")
        elif self._extracting_classmethod():
            result.append("@classmethod\n")

    def _extracting_classmethod(self):
        return self.info.kind == "classmethod"

    def _extracting_staticmethod(self):
        return self.info.kind == "staticmethod"

    def _get_function_signature(self, args):
        args = list(args)
        prefix = ""
        if self._extracting_method() or self._extracting_classmethod():
            self_name = self._get_self_name()
            if self_name is None:
                raise RefactoringError(
                    "Extracting a method from a function with no self argument."
                )
            if self_name in args:
                args.remove(self_name)
            args.insert(0, self_name)
        return prefix + self.info.new_name + "(%s)" % self._get_comma_form(args)

    def _extracting_method(self):
        return not self._extracting_staticmethod() and (
            self.info.method
            and not self.info.make_global
            and _get_function_kind(self.info.scope) == "method"
        )

    def _get_self_name(self):
        if self._extracting_classmethod():
            return "cls"
        return self._get_scope_self_name()

    def _get_scope_self_name(self):
        if self.info.scope.pyobject.get_kind() == "staticmethod":
            return
        param_names = self.info.scope.pyobject.get_param_names()
        if param_names:
            return param_names[0]

    def _get_function_call(self, args):
        return "{prefix}{name}({args})".format(
            prefix=self._get_function_call_prefix(args),
            name=self.info.new_name,
            args=self._get_comma_form(args),
        )

    def _get_function_call_prefix(self, args):
        prefix = ""
        if self.info.method and not self.info.make_global:
            if self._extracting_staticmethod() or self._extracting_classmethod():
                prefix = self.info.scope.parent.pyobject.get_name() + "."
            else:
                self_name = self._get_self_name()
                if self_name in args:
                    args.remove(self_name)
                prefix = self_name + "."
        return prefix

    def _get_comma_form(self, names):
        return ", ".join(names)

    def _get_call(self):
        args = self._find_function_arguments()
        returns = self._find_function_returns()
        call_prefix = ""
        if returns and (not self.info.one_line or self.info.returning_named_expr):
            assignment_operator = " := " if self.info.one_line else " = "
            call_prefix = self._get_comma_form(returns) + assignment_operator
        if self.info.returned:
            call_prefix = "return "
        return call_prefix + self._get_function_call(args)

    def _find_function_arguments(self):
        # if not make_global, do not pass any global names; they are
        # all visible.
        if self.info.global_ and not self.info.make_global:
            return list(
                self.info_collector.read
                & self.info_collector.postread
                & self.info_collector.written
            )
        if not self.info.one_line:
            result = self.info_collector.prewritten & self.info_collector.read
            result |= (
                self.info_collector.prewritten
                & self.info_collector.postread
                & (self.info_collector.maybe_written - self.info_collector.written)
            )
            return list(result)
        start = self.info.region[0]
        if start == self.info.lines_region[0]:
            start = start + re.search("\\S", self.info.extracted).start()
        function_definition = self.info.source[start : self.info.region[1]]
        read = _VariableReadsAndWritesFinder.find_reads_for_one_liners(
            function_definition
        )
        return list(self.info_collector.prewritten.intersection(read))

    def _find_function_returns(self):
        if self.info.one_line:
            written = self.info_collector.written | self.info_collector.maybe_written
            return list(written & self.info_collector.postread)

        if self.info.returned:
            return []
        written = self.info_collector.written | self.info_collector.maybe_written
        return list(written & self.info_collector.postread)

    def _get_unindented_function_body(self, returns):
        if self.info.one_line:
            return self._get_single_expression_function_body()
        return self._get_multiline_function_body(returns)

    def _get_multiline_function_body(self, returns):
        unindented_body = sourceutils.fix_indentation(self.info.extracted, 0)
        unindented_body = self._insert_globals(unindented_body)
        unindented_body = self._insert_nonlocals(unindented_body)
        if returns:
            unindented_body += "\nreturn %s" % self._get_comma_form(returns)
        return unindented_body

    def _get_single_expression_function_body(self):
        extracted = _get_single_expression_body(self.info.extracted, info=self.info)
        body = "return " + extracted
        body = self._insert_globals(body)
        body = self._insert_nonlocals(body)
        return body

    def _insert_globals(self, unindented_body):
        globals_in_body = self._get_globals_in_body(unindented_body)
        globals_ = self.info_collector.globals_ & (
            self.info_collector.written | self.info_collector.maybe_written
        )
        globals_ = globals_ - globals_in_body

        if globals_:
            unindented_body = "global {}\n{}".format(
                ", ".join(globals_), unindented_body
            )
        return unindented_body

    def _insert_nonlocals(self, unindented_body):
        nonlocals_in_body = self._get_nonlocals_in_body(unindented_body)
        nonlocals_ = self.info_collector.nonlocals_ & (
            self.info_collector.written | self.info_collector.maybe_written
        )
        nonlocals_ = nonlocals_ - nonlocals_in_body

        if nonlocals_:
            unindented_body = "nonlocal {}\n{}".format(
                ", ".join(nonlocals_), unindented_body
            )
        return unindented_body

    @staticmethod
    def _get_globals_in_body(unindented_body):
        node = _parse_text(unindented_body)
        visitor = _GlobalFinder()
        visitor.visit(node)
        return visitor.globals_

    @staticmethod
    def _get_nonlocals_in_body(unindented_body):
        node = _parse_text(unindented_body)
        visitor = _NonlocalFinder()
        visitor.visit(node)
        return visitor.nonlocals_


class _ExtractVariableParts:
    def __init__(self, info):
        self.info = info

    def get_definition(self):
        extracted = _get_single_expression_body(self.info.extracted, info=self.info)
        return self.info.new_name + " = " + extracted + "\n"

    def get_body_pattern(self):
        return "(%s)" % self.info.extracted.strip()

    def get_replacement_pattern(self):
        return self.info.new_name

    def get_checks(self):
        return {}


class _FunctionInformationCollector(ast.RopeNodeVisitor):
    def __init__(self, start, end, is_global):
        self.start = start
        self.end = end
        self.is_global = is_global
        self.prewritten = OrderedSet()
        self.maybe_written = OrderedSet()
        self.written = OrderedSet()
        self.read = OrderedSet()
        self.postread = OrderedSet()
        self.postwritten = OrderedSet()
        self.host_function = True
        self.conditional = False
        self.globals_ = OrderedSet()
        self.nonlocals_ = OrderedSet()
        self.surrounded_by_loop = 0
        self.loop_depth = 0

    def _read_variable(self, name, lineno):
        if self.start <= lineno <= self.end:
            if name not in self.written:
                if not self.conditional or name not in self.maybe_written:
                    self.read.add(name)
        if self.end < lineno:
            if name not in self.postwritten:
                self.postread.add(name)

    def _written_variable(self, name, lineno):
        if self.start <= lineno <= self.end:
            if self.conditional:
                self.maybe_written.add(name)
            else:
                self.written.add(name)
            if self.loop_depth > 0 and name in self.read:
                self.postread.add(name)
        if self.start > lineno:
            self.prewritten.add(name)
        if self.end < lineno:
            self.postwritten.add(name)

    def _FunctionDef(self, node):
        if not self.is_global and self.host_function:
            self.host_function = False
            for name in _get_argnames(node.args):
                self._written_variable(name, node.lineno)
            for child in node.body:
                self.visit(child)
        else:
            self._written_variable(node.name, node.lineno)
            visitor = _VariableReadsAndWritesFinder()
            for child in node.body:
                visitor.visit(child)
            for name in visitor.read - visitor.written:
                self._read_variable(name, node.lineno)

    def _Global(self, node):
        self.globals_.add(*node.names)

    def _Nonlocal(self, node):
        self.nonlocals_.add(*node.names)

    def _AsyncFunctionDef(self, node):
        self._FunctionDef(node)

    def _Name(self, node):
        if isinstance(node.ctx, (ast.Store, ast.AugStore)):
            self._written_variable(node.id, node.lineno)
        if not isinstance(node.ctx, ast.Store):
            self._read_variable(node.id, node.lineno)

    def _MatchAs(self, node):
        self._written_variable(node.name, node.lineno)
        if node.pattern:
            self.visit(node.pattern)

    def _Assign(self, node):
        self.visit(node.value)
        for child in node.targets:
            self.visit(child)

    def _AugAssign(self, node):
        self.visit(node.value)
        if isinstance(node.target, ast.Name):
            target_id = node.target.id
            self._read_variable(target_id, node.target.lineno)
            self._written_variable(target_id, node.target.lineno)
        else:
            self.visit(node.target)

    def _ClassDef(self, node):
        self._written_variable(node.name, node.lineno)

    def _ListComp(self, node):
        self._comp_exp(node)

    def _GeneratorExp(self, node):
        self._comp_exp(node)

    def _SetComp(self, node):
        self._comp_exp(node)

    def _DictComp(self, node):
        self._comp_exp(node)

    def _comp_exp(self, node):
        read = OrderedSet(self.read)
        written = OrderedSet(self.written)
        maybe_written = OrderedSet(self.maybe_written)

        for child in ast.iter_child_nodes(node):
            self.visit(child)

        comp_names = list(
            chain.from_iterable(
                self._flatten_nested_tuple_of_names(generator.target)
                for generator in node.generators
            )
        )
        self.read = self.read - comp_names | read
        self.written = self.written - comp_names | written
        self.maybe_written = self.maybe_written - comp_names | maybe_written

    def _flatten_nested_tuple_of_names(self, node):
        if isinstance(node, ast.Tuple):
            for elt in node.elts:
                yield self._flatten_nested_tuple_of_names(elt)
        elif isinstance(node, ast.Name):
            yield node.id
        else:
            assert False, f"Unexpected node type in list comprehension target: {node!r}"

    def _If(self, node):
        self._handle_conditional_node(node)

    def _While(self, node):
        with self._handle_loop_context(node):
            self._handle_conditional_node(node)

    def _For(self, node):
        with self._handle_loop_context(node), self._handle_conditional_context(node):
            # iter has to be checked before the target variables
            self.visit(node.iter)
            self.visit(node.target)

            for child in node.body:
                self.visit(child)
            for child in node.orelse:
                self.visit(child)

    def _handle_conditional_node(self, node):
        with self._handle_conditional_context(node):
            for child in ast.iter_child_nodes(node):
                self.visit(child)

    @contextmanager
    def _handle_conditional_context(self, node):
        if self.start <= node.lineno <= self.end:
            self.conditional = True
        try:
            yield
        finally:
            self.conditional = False

    @contextmanager
    def _handle_loop_context(self, node):
        if node.lineno < self.start:
            self.loop_depth += 1
        try:
            yield
        finally:
            self.loop_depth -= 1


def _get_argnames(arguments):
    result = []
    result.extend(node.arg for node in getattr(arguments, "posonlyargs", []))
    result.extend(node.arg for node in arguments.args)
    if arguments.vararg:
        result.append(arguments.vararg.arg)
    if arguments.kwarg:
        result.append(arguments.kwarg.arg)
    result.extend(node.arg for node in arguments.kwonlyargs)
    return result


class _VariableReadsAndWritesFinder(ast.RopeNodeVisitor):
    def __init__(self):
        self.written = set()
        self.read = set()

    def _Name(self, node):
        if isinstance(node.ctx, (ast.Store, ast.AugStore)):
            self.written.add(node.id)
        if not isinstance(node, ast.Store):
            self.read.add(node.id)

    def _FunctionDef(self, node):
        self.written.add(node.name)
        visitor = _VariableReadsAndWritesFinder()
        for child in ast.iter_child_nodes(node):
            visitor.visit(child)
        self.read.update(visitor.read - visitor.written)

    def _Class(self, node):
        self.written.add(node.name)

    @staticmethod
    def find_reads_and_writes(code):
        if code.strip() == "":
            return set(), set()
        node = _parse_text(code)
        visitor = _VariableReadsAndWritesFinder()
        visitor.visit(node)
        return visitor.read, visitor.written

    @staticmethod
    def find_reads_for_one_liners(code):
        if code.strip() == "":
            return set(), set()
        node = _parse_text(code)
        visitor = _VariableReadsAndWritesFinder()
        visitor.visit(node)
        return visitor.read


class _BaseErrorFinder(ast.RopeNodeVisitor):
    @classmethod
    def has_errors(cls, code):
        if code.strip() == "":
            return False
        node = _parse_text(code)
        visitor = cls()
        visitor.visit(node)
        return visitor.error


class _UnmatchedBreakOrContinueFinder(_BaseErrorFinder):
    def __init__(self):
        self.error = False
        self.loop_count = 0

    def _For(self, node):
        self.loop_encountered(node)

    def _While(self, node):
        self.loop_encountered(node)

    def loop_encountered(self, node):
        self.loop_count += 1
        for child in node.body:
            self.visit(child)
        self.loop_count -= 1
        if node.orelse:
            if isinstance(node.orelse, (list, tuple)):
                for node_ in node.orelse:
                    self.visit(node_)
            else:
                self.visit(node.orelse)

    def _Break(self, node):
        self.check_loop()

    def _Continue(self, node):
        self.check_loop()

    def check_loop(self):
        if self.loop_count < 1:
            self.error = True

    def _FunctionDef(self, node):
        pass

    def _ClassDef(self, node):
        pass


class _AsyncStatementFinder(_BaseErrorFinder):
    def __init__(self):
        self.error = False

    def _AsyncFor(self, node):
        self.error = True

    def _AsyncWith(self, node):
        self.error = True

    def _FunctionDef(self, node):
        pass

    def _ClassDef(self, node):
        pass


class _UnbalancedRegionFinder(_BaseErrorFinder):
    """
    Flag an error if we are including the start of a block without the end.
    We detect this by ensuring there is no AST node that starts inside the
    selected range but ends outside of it.
    """

    def __init__(self, line_start: int, line_end: int):
        self.error = False
        self.line_start = line_start
        self.line_end = line_end

    def generic_visit(self, node: ast.AST):
        if not hasattr(node, "end_lineno"):
            super().generic_visit(node)  # Visit children
            return
        ends_before_range_starts = node.end_lineno < self.line_start
        starts_after_range_ends = node.lineno > self.line_end
        if ends_before_range_starts or starts_after_range_ends:
            return  # Don't visit children
        starts_on_or_after_range_start = node.lineno >= self.line_start
        ends_after_range_ends = node.end_lineno > self.line_end
        if starts_on_or_after_range_start and ends_after_range_ends:
            self.error = True
            return  # Don't visit children
        super().generic_visit(node)  # Visit children


class _GlobalFinder(ast.RopeNodeVisitor):
    def __init__(self):
        self.globals_ = OrderedSet()

    def _Global(self, node):
        self.globals_.add(*node.names)


class _NonlocalFinder(ast.RopeNodeVisitor):
    def __init__(self):
        self.nonlocals_ = OrderedSet()

    def _Nonlocal(self, node):
        self.nonlocals_.add(*node.names)


def _get_function_kind(scope):
    return scope.pyobject.get_kind()


def _parse_text(body):
    body = sourceutils.fix_indentation(body, 0)
    try:
        node = ast.parse(body)
    except SyntaxError:
        # needed to parse expression containing := operator
        try:
            node = ast.parse("(" + body + ")")
        except SyntaxError:
            node = ast.parse(
                "async def __rope_placeholder__():\n"
                + sourceutils.fix_indentation(body, 4)
            )
            node.body = node.body[0].body
    return node


def _join_lines(code):
    lines = []
    for line in code.splitlines():
        if line.endswith("\\"):
            lines.append(line[:-1].strip())
        else:
            lines.append(line.strip())
    return " ".join(lines)


def _get_single_expression_body(extracted, info):
    extracted = sourceutils.fix_indentation(extracted, 0)
    already_parenthesized = (
        extracted.lstrip()[0] in "({[" and extracted.rstrip()[-1] in ")}]"
    )
    large_multiline = extracted.count("\n") >= 2 and already_parenthesized
    if not large_multiline:
        extracted = _join_lines(extracted)
    multiline_expression = "\n" in extracted
    if (
        info.returning_named_expr
        or info.returning_generator_exp
        or (multiline_expression and not large_multiline)
    ):
        extracted = "(" + extracted + ")"
    return extracted



from typing import List, Union

from rope.base import ast, exceptions, pynames, pynamesdef, utils
from rope.refactor.importutils import actions, importinfo


class ModuleImports:
    def __init__(self, project, pymodule, import_filter=None):
        self.project = project
        self.pymodule = pymodule
        self.separating_lines = 0
        self.filter = import_filter
        self.sorted = False

    @property
    @utils.saveit
    def imports(self):
        finder = _GlobalImportFinder(self.pymodule)
        result = finder.find_import_statements()
        self.separating_lines = finder.get_separating_line_count()
        if self.filter is not None:
            for import_stmt in result:
                if not self.filter(import_stmt):
                    import_stmt.readonly = True
        return result

    def _get_unbound_names(self, defined_pyobject):
        visitor = _GlobalUnboundNameFinder(self.pymodule, defined_pyobject)
        visitor.visit(self.pymodule.get_ast())
        return visitor.unbound

    def _get_all_star_list(self, pymodule):
        def _resolve_name(
            name: Union[pynamesdef.AssignedName, pynames.ImportedName]
        ) -> List:
            while isinstance(name, pynames.ImportedName):
                try:
                    name = name.imported_module.get_object().get_attribute(
                        name.imported_name,
                    )
                except exceptions.AttributeNotFoundError:
                    return []
            assert isinstance(name, pynamesdef.AssignedName)
            return name.assignments

        result = set()
        try:
            all_star_list = pymodule.get_attribute("__all__")
        except exceptions.AttributeNotFoundError:
            return result

        assignments = [
            assignment.ast_node for assignment in _resolve_name(all_star_list)
        ]

        # FIXME: Need a better way to recursively infer possible values.
        #        Currently pyobjects can recursively infer type, but not values.
        # Do a very basic 1-level value inference

        while assignments:
            assignment = assignments.pop()
            if isinstance(assignment, ast.List):
                stack = list(assignment.elts)
                while stack:
                    el = stack.pop()
                    if isinstance(el, ast.IfExp):
                        stack.append(el.body)
                        stack.append(el.orelse)
                    elif isinstance(el, ast.Starred):
                        assignments.append(el.value)
                    else:
                        if isinstance(el, ast.Constant) and isinstance(el.value, str):
                            result.add(el.value)
                        elif isinstance(el, ast.Name):
                            try:
                                name = pymodule.get_attribute(el.id)
                            except exceptions.AttributeNotFoundError:
                                continue
                            else:
                                for av in _resolve_name(name):
                                    if isinstance(
                                        av.ast_node,
                                        ast.Constant,
                                    ) and isinstance(
                                        av.ast_node.value,
                                        str,
                                    ):
                                        result.add(av.ast_node.value)
            elif isinstance(assignment, ast.Name):
                try:
                    name = pymodule.get_attribute(assignment.id)
                except exceptions.AttributeNotFoundError:
                    continue
                else:
                    assignments.extend(
                        assignment.ast_node for assignment in _resolve_name(name)
                    )
            elif isinstance(assignment, ast.BinOp):
                assignments.append(assignment.left)
                assignments.append(assignment.right)
        return result

    def remove_unused_imports(self):
        can_select = _OneTimeSelector(
            self._get_unbound_names(self.pymodule)
            | self._get_all_star_list(self.pymodule)
            | {"__all__"}
        )
        visitor = actions.RemovingVisitor(
            self.project, self._current_folder(), can_select
        )
        for import_statement in self.imports:
            import_statement.accept(visitor)

    def get_used_imports(self, defined_pyobject):
        result = []
        can_select = _OneTimeSelector(self._get_unbound_names(defined_pyobject))
        visitor = actions.FilteringVisitor(
            self.project, self._current_folder(), can_select
        )
        for import_statement in self.imports:
            new_import = import_statement.accept(visitor)
            if new_import is not None and not new_import.is_empty():
                result.append(new_import)
        return result

    def get_changed_source(self):
        if not self.project.prefs.get("pull_imports_to_top") and not self.sorted:
            return "".join(self._rewrite_imports(self.imports))

        # Make sure we forward a removed import's preceding blank
        # lines count to the following import statement.
        prev_stmt = None
        for stmt in self.imports:
            if prev_stmt is not None and prev_stmt.import_info.is_empty():
                stmt.blank_lines = max(prev_stmt.blank_lines, stmt.blank_lines)
            prev_stmt = stmt
        # The new list of imports.
        imports = [stmt for stmt in self.imports if not stmt.import_info.is_empty()]

        after_removing = self._remove_imports(self.imports)
        first_non_blank = self._first_non_blank_line(after_removing, 0)
        first_import = self._first_import_line() - 1
        result = []
        # Writing module docs
        result.extend(after_removing[first_non_blank:first_import])
        # Writing imports
        sorted_imports = sorted(imports, key=self._get_location)
        for stmt in sorted_imports:
            if stmt != sorted_imports[0]:
                result.append("\n" * stmt.blank_lines)
            result.append(stmt.get_import_statement() + "\n")
        if sorted_imports and first_non_blank < len(after_removing):
            result.append("\n" * self.separating_lines)

        # Writing the body
        first_after_imports = self._first_non_blank_line(after_removing, first_import)
        result.extend(after_removing[first_after_imports:])
        return "".join(result)

    def _get_import_location(self, stmt):
        start = stmt.get_new_start()
        if start is None:
            start = stmt.get_old_location()[0]
        return start

    def _get_location(self, stmt):
        if stmt.get_new_start() is not None:
            return stmt.get_new_start()
        else:
            return stmt.get_old_location()[0]

    def _remove_imports(self, imports):
        lines = self.pymodule.source_code.splitlines(True)
        after_removing = []
        first_import_line = self._first_import_line()
        last_index = 0
        for stmt in imports:
            start, end = stmt.get_old_location()
            blank_lines = 0
            if start != first_import_line:
                blank_lines = _count_blank_lines(
                    lines.__getitem__, start - 2, last_index - 1, -1
                )
            after_removing.extend(lines[last_index : start - 1 - blank_lines])
            last_index = end - 1
        after_removing.extend(lines[last_index:])
        return after_removing

    def _rewrite_imports(self, imports):
        lines = self.pymodule.source_code.splitlines(True)
        after_rewriting = []
        last_index = 0
        for stmt in imports:
            start, end = stmt.get_old_location()
            after_rewriting.extend(lines[last_index : start - 1])
            if not stmt.import_info.is_empty():
                after_rewriting.append(stmt.get_import_statement() + "\n")
            last_index = end - 1
        after_rewriting.extend(lines[last_index:])
        return after_rewriting

    def _first_non_blank_line(self, lines, lineno):
        return lineno + _count_blank_lines(lines.__getitem__, lineno, len(lines))

    def add_import(self, import_info):
        visitor = actions.AddingVisitor(self.project, [import_info])
        for import_statement in self.imports:
            if import_statement.accept(visitor):
                break
        else:
            lineno = self._get_new_import_lineno()
            blanks = self._get_new_import_blanks()
            self.imports.append(
                importinfo.ImportStatement(
                    import_info, lineno, lineno, blank_lines=blanks
                )
            )

    def _get_new_import_blanks(self):
        return 0

    def _get_new_import_lineno(self):
        if self.imports:
            return self.imports[-1].end_line
        return 1

    def filter_names(self, can_select):
        visitor = actions.RemovingVisitor(
            self.project, self._current_folder(), can_select
        )
        for import_statement in self.imports:
            import_statement.accept(visitor)

    def expand_stars(self):
        can_select = _OneTimeSelector(self._get_unbound_names(self.pymodule))
        visitor = actions.ExpandStarsVisitor(
            self.project, self._current_folder(), can_select
        )
        for import_statement in self.imports:
            import_statement.accept(visitor)

    def remove_duplicates(self):
        added_imports = []
        for import_stmt in self.imports:
            visitor = actions.AddingVisitor(self.project, [import_stmt.import_info])
            for added_import in added_imports:
                if added_import.accept(visitor):
                    import_stmt.empty_import()
            else:
                added_imports.append(import_stmt)

    def force_single_imports(self):
        """force a single import per statement"""
        for import_stmt in self.imports[:]:
            import_info = import_stmt.import_info
            if import_info.is_empty() or import_stmt.readonly:
                continue
            if len(import_info.names_and_aliases) > 1:
                for name_and_alias in import_info.names_and_aliases:
                    if hasattr(import_info, "module_name"):
                        new_import = importinfo.FromImport(
                            import_info.module_name, import_info.level, [name_and_alias]
                        )
                    else:
                        new_import = importinfo.NormalImport([name_and_alias])
                    self.add_import(new_import)
                import_stmt.empty_import()

    def get_relative_to_absolute_list(self):
        visitor = actions.RelativeToAbsoluteVisitor(
            self.project, self._current_folder()
        )
        for import_stmt in self.imports:
            if not import_stmt.readonly:
                import_stmt.accept(visitor)
        return visitor.to_be_absolute

    def get_self_import_fix_and_rename_list(self):
        visitor = actions.SelfImportVisitor(
            self.project, self._current_folder(), self.pymodule.get_resource()
        )
        for import_stmt in self.imports:
            if not import_stmt.readonly:
                import_stmt.accept(visitor)
        return visitor.to_be_fixed, visitor.to_be_renamed

    def _current_folder(self):
        return self.pymodule.get_resource().parent

    def sort_imports(self):
        if self.project.prefs.get("sort_imports_alphabetically"):
            sort_kwargs = dict(key=self._get_import_name)
        else:
            sort_kwargs = dict(key=self._key_imports)

        # IDEA: Sort from import list
        visitor = actions.SortingVisitor(self.project, self._current_folder())
        for import_statement in self.imports:
            import_statement.accept(visitor)
        in_projects = sorted(visitor.in_project, **sort_kwargs)
        third_party = sorted(visitor.third_party, **sort_kwargs)
        standards = sorted(visitor.standard, **sort_kwargs)
        future = sorted(visitor.future, **sort_kwargs)
        last_index = self._first_import_line()
        last_index = self._move_imports(future, last_index, 0)
        last_index = self._move_imports(standards, last_index, 1)
        last_index = self._move_imports(third_party, last_index, 1)
        last_index = self._move_imports(in_projects, last_index, 1)
        self.separating_lines = 2
        self.sorted = True

    def _first_import_line(self):
        nodes = self.pymodule.get_ast().body
        lineno = 0
        if self.pymodule.get_doc() is not None:
            lineno = 1
        if len(nodes) > lineno:
            if isinstance(nodes[lineno], ast.Import) or isinstance(
                nodes[lineno], ast.ImportFrom
            ):
                return nodes[lineno].lineno
            first_line = get_first_decorator_or_function_start_line(nodes[lineno])
            lineno = self.pymodule.logical_lines.logical_line_in(first_line)[0]
        else:
            lineno = self.pymodule.lines.length()

        return lineno - _count_blank_lines(
            self.pymodule.lines.get_line, lineno - 1, 1, -1
        )

    def _get_import_name(self, import_stmt):
        import_info = import_stmt.import_info
        if hasattr(import_info, "module_name"):
            return "{}.{}".format(
                import_info.module_name,
                import_info.names_and_aliases[0][0],
            )
        else:
            return import_info.names_and_aliases[0][0]

    def _key_imports(self, stm1):
        str1 = stm1.get_import_statement()
        return str1.startswith("from "), str1

        # str1 = stmt1.get_import_statement()
        # str2 = stmt2.get_import_statement()
        # if str1.startswith('from ') and not str2.startswith('from '):
        #    return 1
        # if not str1.startswith('from ') and str2.startswith('from '):
        #    return -1
        # return cmp(str1, str2)

    def _move_imports(self, imports, index, blank_lines):
        if imports:
            imports[0].move(index, blank_lines)
            index += 1
            if len(imports) > 1:
                for stmt in imports[1:]:
                    stmt.move(index)
                    index += 1
        return index

    def handle_long_imports(self, maxdots, maxlength):
        visitor = actions.LongImportVisitor(
            self._current_folder(), self.project, maxdots, maxlength
        )
        for import_statement in self.imports:
            if not import_statement.readonly:
                import_statement.accept(visitor)
        for import_info in visitor.new_imports:
            self.add_import(import_info)
        return visitor.to_be_renamed

    def remove_pyname(self, pyname):
        """Removes pyname when imported in ``from mod import x``"""
        visitor = actions.RemovePyNameVisitor(
            self.project, self.pymodule, pyname, self._current_folder()
        )
        for import_stmt in self.imports:
            import_stmt.accept(visitor)


def get_first_decorator_or_function_start_line(node):
    decorators = getattr(node, "decorator_list", [])
    first_line = min([decorator.lineno for decorator in decorators] + [node.lineno])
    return first_line


def _count_blank_lines(get_line, start, end, step=1):
    count = 0
    for idx in range(start, end, step):
        if get_line(idx).strip() == "":
            count += 1
        else:
            break
    return count


class _OneTimeSelector:
    def __init__(self, names):
        self.names = names
        self.selected_names = set()

    def __call__(self, imported_primary):
        if self._can_name_be_added(imported_primary):
            for name in self._get_dotted_tokens(imported_primary):
                self.selected_names.add(name)
            return True
        return False

    def _get_dotted_tokens(self, imported_primary):
        tokens = imported_primary.split(".")
        for i in range(len(tokens)):
            yield ".".join(tokens[: i + 1])

    def _can_name_be_added(self, imported_primary):
        for name in self._get_dotted_tokens(imported_primary):
            if name in self.names and name not in self.selected_names:
                return True
        return False


class _UnboundNameFinder(ast.RopeNodeVisitor):
    def __init__(self, pyobject):
        self.pyobject = pyobject

    def _visit_child_scope(self, node):
        pyobject = (
            self.pyobject.get_module()
            .get_scope()
            .get_inner_scope_for_line(node.lineno)
            .pyobject
        )
        visitor = _LocalUnboundNameFinder(pyobject, self)
        for child in ast.iter_child_nodes(node):
            visitor.visit(child)

    def _FunctionDef(self, node):
        self._visit_child_scope(node)

    def _ClassDef(self, node):
        self._visit_child_scope(node)

    def _Name(self, node):
        if self._get_root()._is_node_interesting(node) and not self.is_bound(node.id):
            self.add_unbound(node.id)

    def _Attribute(self, node):
        result = []
        while isinstance(node, ast.Attribute):
            result.append(node.attr)
            node = node.value
        if isinstance(node, ast.Name):
            result.append(node.id)
            primary = ".".join(reversed(result))
            if self._get_root()._is_node_interesting(node) and not self.is_bound(
                primary
            ):
                self.add_unbound(primary)
        else:
            self.visit(node)

    def _get_root(self):
        pass

    def is_bound(self, name, propagated=False):
        pass

    def add_unbound(self, name):
        pass


class _GlobalUnboundNameFinder(_UnboundNameFinder):
    def __init__(self, pymodule, wanted_pyobject):
        super().__init__(pymodule)
        self.unbound = set()
        self.names = set()
        for name, pyname in pymodule._get_structural_attributes().items():
            if not isinstance(pyname, (pynames.ImportedName, pynames.ImportedModule)):
                self.names.add(name)
        wanted_scope = wanted_pyobject.get_scope()
        self.start = wanted_scope.get_start()
        self.end = wanted_scope.get_end() + 1

    def _get_root(self):
        return self

    def is_bound(self, primary, propagated=False):
        name = primary.split(".")[0]
        return name in self.names

    def add_unbound(self, name):
        names = name.split(".")
        for i in range(len(names)):
            self.unbound.add(".".join(names[: i + 1]))

    def _is_node_interesting(self, node):
        return self.start <= node.lineno < self.end


class _LocalUnboundNameFinder(_UnboundNameFinder):
    def __init__(self, pyobject, parent):
        super().__init__(pyobject)
        self.parent = parent

    def _get_root(self):
        return self.parent._get_root()

    def is_bound(self, primary, propagated=False):
        name = primary.split(".")[0]
        if propagated:
            names = self.pyobject.get_scope().get_propagated_names()
        else:
            names = self.pyobject.get_scope().get_names()
        return name in names or self.parent.is_bound(name, propagated=True)

    def add_unbound(self, name):
        self.parent.add_unbound(name)


class _GlobalImportFinder:
    def __init__(self, pymodule):
        self.current_folder = None
        if pymodule.get_resource():
            self.current_folder = pymodule.get_resource().parent
            self.pymodule = pymodule
        self.imports = []
        self.pymodule = pymodule
        self.lines = self.pymodule.lines

    def visit_import(self, node, end_line):
        start_line = node.lineno
        import_statement = importinfo.ImportStatement(
            importinfo.NormalImport(self._get_names(node.names)),
            start_line,
            end_line,
            self._get_text(start_line, end_line),
            blank_lines=self._count_empty_lines_before(start_line),
        )
        self.imports.append(import_statement)

    def _count_empty_lines_before(self, lineno):
        return _count_blank_lines(self.lines.get_line, lineno - 1, 0, -1)

    def _count_empty_lines_after(self, lineno):
        return _count_blank_lines(self.lines.get_line, lineno + 1, self.lines.length())

    def get_separating_line_count(self):
        if not self.imports:
            return 0
        return self._count_empty_lines_after(self.imports[-1].end_line - 1)

    def _get_text(self, start_line, end_line):
        result = [self.lines.get_line(index) for index in range(start_line, end_line)]
        return "\n".join(result)

    def visit_from(self, node, end_line):
        level = 0
        if node.level:
            level = node.level
        import_info = importinfo.FromImport(
            node.module or "",
            level,
            self._get_names(node.names),
        )
        start_line = node.lineno
        self.imports.append(
            importinfo.ImportStatement(
                import_info,
                node.lineno,
                end_line,
                self._get_text(start_line, end_line),
                blank_lines=self._count_empty_lines_before(start_line),
            )
        )

    def _get_names(self, alias_names):
        return [(alias.name, alias.asname) for alias in alias_names]

    def find_import_statements(self):
        nodes = self.pymodule.get_ast().body
        for node in nodes:
            if isinstance(node, (ast.Import, ast.ImportFrom)):
                lines = self.pymodule.logical_lines
                end_line = lines.logical_line_in(node.lineno)[1] + 1
            if isinstance(node, ast.Import):
                self.visit_import(node, end_line)
            if isinstance(node, ast.ImportFrom):
                self.visit_from(node, end_line)
        return self.imports
